importPackage(java.lang);
importPackage(java.sql);
importPackage(java.util);
importPackage(java.io);
importPackage(java.net);
importPackage(Packages.javax.xml);
importPackage(Packages.javax.xml.parsers);
importPackage(Packages.javax.xml.transform);
importPackage(Packages.javax.xml.transform.dom);
importPackage(Packages.javax.xml.transform.stream);
importPackage(Packages.org.w3c.dom);
importPackage(Packages.com.alarmpoint.servicemanager.cm);
importClass(Packages.com.alarmpoint.integrationagent.apxml.APXMLMessage);
importClass(Packages.com.alarmpoint.integrationagent.apxml.APXMLToken);
importClass(Packages.com.thoughtworks.xstream.XStream);
importClass(Packages.com.thoughtworks.xstream.converters.reflection.PureJavaReflectionProvider);

var xStream = new XStream(new PureJavaReflectionProvider())


var smUrl = "http://localhost:13080/sc62server/ws";
var exceptionString = "JavaException: com.alarmpoint.servicemanager.cm.ServiceManagerException: ";

//only used for Callout Annotations from the CALLOUT APS script
var calloutAnnotateUser = "falcon";
var calloutAnnotatePass = "";
var notePrefix = ""; // The Prefix which is added to the note comments for AlarmPoint e.g. "[AlarmPoint] " 

var TOKEN_NAME = "apiaResult";
var TOKEN_NAME_EXCEPTION = "apiaResultException";

var smConnection = null;

/**
 * This is the main method for handling APXML messages sent to the
 * Integration Service by APClient.  The messages may be requests to perform
 * local activity, or they may be requests to make submissions to AlarmPoint.
 * <p>
 * Any APXMLMessage object that this method returns will be sent to AlarmPoint
 * via the Integration Service's outbound queues.
 */
function apia_input(apxml)
{
  return apxml;
}

/**
 * This is the main method for handling APXML messages sent to the
 * Integration Service by AlarmPoint.  The messages may be responses to
 * previous submissions, or they may be requests to the Integration Service
 * that originate from AlarmPoint
 * (e.g., via an ExternalServiceMessage/ExternalServiceRequest2).
 * <p>
 * Any APXMLMessage object that this method returns will be sent to AlarmPoint
 * either via the Integration Service's outbound queues or directly as a
 * Web Service response, depending on the mechanism used to submit the APXML.
 */
function apia_response(apxml)
{
  var method = apxml.getMethod();
  if (APXMLMessage.OK_METHOD.equalsIgnoreCase(method))
  {
    return handleOK(apxml);
  }
  else if (APXMLMessage.ERROR_METHOD.equalsIgnoreCase(method))
  {
    return handleERROR(apxml);
  }
  else if (APXMLMessage.SEND_METHOD.equalsIgnoreCase(method))
  {
    return handleSend(apxml);
  }
  else if (APXMLMessage.REQUEST_METHOD.equalsIgnoreCase(method))
  {
    return handleRequest(apxml);
  }
  else
  {
    throw new IllegalArgumentException("Unrecognized APXML method: " + method);
  }
}


/**
 * OK responses are returned by AlarmPoint for each APXML message that it accepts
 * from an Integration Service.
 * <p>
 * This implementation just logs the ack.
 */
function handleOK(apxml)
{
  var id = apxml.getTransactionId();
  logInfo("Submission of transaction [" + id + "] succeeded.");
}

/**
 * ERROR responses are returned by AlarmPoint for each APXML message that it
 * is unable to accept from an Integration Service.
 * <p>
 * This implementation just logs the error.
 */
function handleERROR(apxml)
{
  var transaction = apxml.getTransactionId();
  var code = apxml.getValue(APXMLMessage.ERROR_CODE);
  ServiceAPI.getLogger().error("Submission of transaction " + transaction + "failed with error code " + code);
}

/**
 * This implementation just logs query request, there shouldn't be any direct calls to this method
 */
function handleRequest(apxml)
{
  return handleSend(apxml);
}



/**
 * The main routine for handling Requests AlarmPoint
 * <p>
 * Throw an exception that will be logged by the Integration Agent.
 */
function handleSend(apxml)
{
  logDebug( "handleSend received apxml: " + apxml ); 

  // handle annotations coming from callout APS script 
  // they are formatted differently 
  // the presence of the message token means this message 
  // is a callout annotation. 
  var message = apxml.getValue("message");
  if ( message != null )
  {
    var incidentId = apxml.getValue("incident_id");
    return handleCalloutAnnotate( incidentId, message );
  }
  else
  {
    return handleApsResponses( apxml );
  }
}
  
  
/**
 * The main routine for handling Requests from AlarmPoint HP SM CM APS
 */  
function handleApsResponses( apxml )
{
  var response = initializeResponse(apxml);
  
  // message fields
  var requestText = apxml.getValue("request_text");
  var messageText = apxml.getValue("message_text");
  var result = null;
 
  var smUser = null;
  var smPass = null;
  var changeId = null;

  try
  {
	  // deserialize the additional tokens
	  if (apxml.getValue("additionalTokens") != null && apxml.getValue("additionalTokens") != "") 
	  {
	    var additionalTokens = xStream.fromXML(apxml.getValue("additionalTokens")); // LinkedHashMap
	    
	    logDebug( "additionalTokens keys: " + additionalTokens);
	    smUser = processAdditionalTokens(additionalTokens, "sm_user", true); 
	    smPass = processAdditionalTokens(additionalTokens, "sm_pass", true);
	    changeId = processAdditionalTokens(additionalTokens, "change_id", true); 
	  }
	  else
	  {
	    throw new Exception("additionalTokens is missing from apxml"); 
	  }

	  logDebug("smUser: " + smUser + " smPass: " + smPass + " changeId: " + changeId);
	  logDebug("incident_id: " + changeId + " requestText: " + requestText);
	  logDebug("messageText: " + messageText);

	  smConnection = new ServiceManagerScriptObject( smUrl, smUser, smPass );
    
	  if (messageText != null && messageText != "")
	  {
	    messageText = notePrefix + messageText;
	  } 
	  else 
	  {
	    messageText = notePrefix + "Updated Change";
	  }

	  logDebug("About to call Service Manager requestText: [" + requestText + "]");

      if (requestText == "annotate" || requestText == "ignore") 
	  {
	    result = smConnection.annotateChange(changeId, messageText);
	  }
	  else if (requestText == "approve" )
	  {
		result = smConnection.approveChange(changeId, messageText);
	  }
	  else if (requestText == "retract" )
	  {
		result = smConnection.retractChange(changeId, messageText);
	  }
	  else if (requestText == "deny") 
	  {
		result = smConnection.denyChange(changeId, messageText);
	  }
	  else
	  { 
	    throw new Exception("Request Script Failed: RequestText: [" + requestText + "] not implemented"); 
	  }
      response.setToken(TOKEN_NAME, processResponse(result), APXMLToken.Type.STRING);
  }
  catch (e)
  {
    logError("Caught exception processing [" + requestText + "] Exception:" + e);
    response.setToken(TOKEN_NAME_EXCEPTION, e.toString().replace(exceptionString, ""), APXMLToken.Type.STRING);
  }
  
  logDebug("Finished call to Service Manager for requestText: [" + requestText + "]");
  logDebug("Response: " + response);
  return response;
}

/**
 * handles annotations sent from the Callout APS script
 */
function handleCalloutAnnotate( incidentId, message )
{
  smConnection = ServiceManagerScriptObject( smUrl, calloutAnnotateUser, calloutAnnotatePass );
  
  // incidentId is of the form C109876_Approval_7898 or C109876_Change_Logging 
  // so to get the changeId we need to get the stirng up to the _ 
  var pos = incidentId.indexOf( "_" );
  var changeId; 
  if ( pos > 0 )
  {
    changeId = incidentId.substring( 0, pos );
  }
  else 
  {
    changeId = incidentId; 
  }

  logDebug( "About to call to Service Manager for callout annotate for [" + changeId + "]" );

  try 
  {
    smConnection.annotateChange( changeId, message );
  }
  catch (e)
  {
    logError("Caught exception processing annotation for [" + changeId + "] Exception:" + e);
  }
  
  logDebug("Finished call to Service Manager for callout annotate");
}

/**
 * Initializes the response to be returned to the device
 * @param apxml APXML containing the request information
 * @return initialized instance of the response APXML
 */
function initializeResponse(apxml)
{
  var response = ServiceAPI.createAPXML();
  response.setMethod(APXMLMessage.RESPONSE_METHOD);
  response.setSubclass(APXMLMessage.RESPONSE_SUBCLASS);
  
  response.setToken(APXMLMessage.REQUEST_ID, apxml.getValue(APXMLMessage.REQUEST_ID), APXMLToken.Type.STRING);
  response.setToken(APXMLMessage.DESTINATION, apxml.getValue(APXMLMessage.ORIGINATOR), APXMLToken.Type.STRING);

  if (apxml.hasToken(APXMLMessage.APIA_PRIORITY)) 
  {
    response.setToken(APXMLMessage.APIA_PRIORITY, apxml.getValue(APXMLMessage.APIA_PRIORITY));
  }
  else
  {
    response.setToken(APXMLMessage.APIA_PRIORITY, APXMLMessage.APIA_PRIORITY_HIGH);
  }
  return response;
}

/**
 * Retrieves the additional tokens from the APXML and throws an exception if a mandatory token is missing
 * @param additionalTokens collection of tokens to process
 * @param tokenName name of the token to retrieve
 * @param mandatory true if token is required, false otherwise
 * @return the token value
 */
function processAdditionalTokens(additionalTokens, tokenName, mandatory)
{
	var token = additionalTokens.get(tokenName);
	if (mandatory && token == null)
	{
		throw new Exception("additionalTokens in apxml is missing " + tokenName + " which is mandatory");
	}
	return token;
}

/**
 * Takes the result of the request and processes any messages retrieved into a string to send back in the response
 * @param result Collection of messages from the request
 * @return Updated response object
 */
function processResponse(result)
{
	// As the messages returned from Service Manager will be returned as a list of messages, these need to be processed and written back to the response APXML
	var msg = "";
	if (result != null)
	{
		for (var i=0; i<result.size(); i++)
		{
			msg += result.get(i) + "\n";
		}
	}
	return msg;
}

/**
 * Retrieves the first change that matches the specified parameters.
 *
 * Parameters: a java.util.Map<String, String> of query fields and associated values
 * Return: A com.alarmpoint.servicemanager.ServiceManagerChange object for the first matched change or a String if a error occurred
 */
function getChange(params)
{
  if(params == null)
    return "No search parameters specified";
  
  //This search should only ever return 1 result
  var changes = getChanges(params);
  
  if(typeof changes == "string")
    return changes;
    
  if(changes.isEmpty())
    return "No Changes Returned";
    
  return changes.get(0);
}

/**
 * Retrieves a list of changes that match the specified parameters
 *
 * Parameters: a java.util.Map<String, String> containing a queryObject field and associated ServiceManagerQuery object or 
 *             a queryString field and associated SM query string or 
 *             query fields and associated values 
 * Return: A java.util.Collection<ServiceManagerChange> object containing all matching changes or a String if a error occurred
 */
function getChanges(params) 
{  
  var sm_user = params.remove("smUser");
  var sm_password = params.remove("smPass");
  var client = new ServiceManagerScriptObject(smUrl, sm_user, sm_password);
  if(params == null)
    return "No search parameters specified";
  var query = "";
  if(params.containsKey("queryObject")){
    query = params.get("queryObject");
  }else if(params.containsKey("queryString")){ 
    query = params.get("queryString");
  }else{ 
    query = new ServiceManagerQuery();
    query.constructFromMap(params);
  }
  try{
    return client.queryChange(query);
  }catch(e){
    return e.toString().replace(exceptionString, "");
  }
}

/**
 * Saves the changes made to a Service Manager Change
 *
 * Parameters: a java.util.Map<String, Object> containing a change field and associated ServiceManagerChange object and a changeID field containing the id of the change to update
 * Returns: null if the update was successful or a String if the update failed
 */
function saveChange(params) 
{  
  var sm_user = params.remove("smUser");
  var sm_password = params.remove("smPass");
  var client = new ServiceManagerScriptObject(smUrl, sm_user, sm_password);
  var change = params.get("change");
  var changeID = params.get("changeID");
  
  try{
    var update = change.constructUpdateInstance();
    if(update.get("IMTicketStatus") != "Closed")
      client.updateChange(changeID, change.constructUpdateInstance());
    else
      client.closeChange(changeID, change.constructUpdateInstance());
    return null;
  }catch(e){
    return e.toString().replace(exceptionString, "");
  }
}

/**
 * Adds a activity text row to Service Manager Change
 *
 * Parameters: String activityText changeID field containing the id of the change to update
 * Returns: null if the update was successful or a String if the update failed
 */
function addActivity(params) 
{  
  var sm_user = params.remove("smUser");
  var sm_password = params.remove("smPass");
  var client = new ServiceManagerScriptObject(smUrl, sm_user, sm_password);
  var changeID = params.get("changeID");
  var activityText = params.get("activityText");
  
  try{
    client.annotateChange(changeID, activityText);
    return null;
  }catch(e){
    return e.toString().replace(exceptionString, "");
  }
}

/**
 * Gets the requested date for the selected Service Manager Change
 * @param params collection that contains the changeID field containing the id of the change and the action
 * that will be called via the webservice
 * @return set of ServiceManagerChange records or a String if failed
 */
function getChangeRecords(params)
{
	  var sm_user = params.remove("smUser");
	  var sm_password = params.remove("smPass");
	  var client = new ServiceManagerScriptObject(smUrl, sm_user, sm_password);
	  var changeID = params.get("changeID");
	  var action = params.get("action");
	  
	  try
	  {
	    return client.getChangeRecords(changeID, action, "ChangeNumber");
	  }
	  catch(e)
	  {
	    return e.toString().replace(exceptionString, "");
	  }
}

/**
 * Gets the requested approval records for the selected Service Manager Change
 * 
 * Needed because Changes use "ChangeID" and ChangeApprovals use "ChangeNumber" 
 *    
 * @param params collection that contains the changeID field containing the id of the change and the action
 * that will be called via the webservice
 * @return set of ServiceManagerChange records or a String if failed
 */
function getChangeApprovalRecords(params)
{
	  var sm_user = params.remove("smUser");
	  var sm_password = params.remove("smPass");
	  var client = new ServiceManagerScriptObject(smUrl, sm_user, sm_password);
	  var changeID = params.get("changeID");
	  var action = params.get("action");
	  
	  try
	  {
	    return client.getChangeRecords(changeID, action, "ChangeNumber");
	  }
	  catch(e)
	  {
	    return e.toString().replace(exceptionString, "");
	  }
}

/**
/**
 * Created a new Service Manager Change
 *
 * Parameters: a java.util.Map<String, Object> containing a change field and associated ServiceManagerChange object
 * Returns: A String[] array with the first entry being the id of the new change created (null if the create failed) and the second entry being the error if the create failed (null if the create succeeded)
 */
function createChange(params) 
{  
  var sm_user = params.remove("smUser");
  var sm_password = params.remove("smPass");
  var client = new ServiceManagerScriptObject(smUrl, sm_user, sm_password);
  var change = params.get("change");
  
  var status = java.lang.reflect.Array.newInstance(java.lang.String, 2);
  try{
    status[0] = client.createChange(change.constructUpdateInstance());
    status[1] = null;
    return status;
  }catch(e){
    status[0] = null;
    status[1] = e.toString().replace(exceptionString, "");
    return status;
  }
}

/**
 * Retrieves list values for Category, Subcategory, ProductType and ProblemType predicates.  These lists can either be dynamic lists or static lists
 *
 * Parameters: a java.util.Map<String, boolean>
 * Returns: a java.util.Map<String, Object> containing dynamic lists if the useDynamicLists field is true and static lists otherwise
 */
function getLists(params)
{
  var sm_user = params.remove("smUser");
  var sm_password = params.remove("smPass");
  var client = new ServiceManagerScriptObject(smUrl, sm_user, sm_password);

  try{
    var fields = new HashMap();
    
    fields.put("ClosureCode", client.getSortedList("RetrieveProbcauseListRequest", "cause.code <> \"\"", "cause.code"));
    
    return fields;
  }catch(e){
    return e.toString().replace(exceptionString, "");
  }
}

/**
 * Retrieves a list of objects specified by the Soap operation name and the query
 *
 * Parameters: a java.util.Map containing a operation field and associated SM Soap operation and a query field and associated SM query
 * Returns: a java.util.Collection<Map<String, String>> containing a list of all the objects that matched the query for the operation and for each object a map of field names to values
 */
function getValueMap(params)
{
  var sm_user = params.remove("smUser");
  var sm_password = params.remove("smPass");
  var client = new ServiceManagerScriptObject(smUrl, sm_user, sm_password);
  var operation = params.get("operation");
  var query = params.get("query");
  
  try{
    return client.getValueMap(operation, query); 
  }catch(e){
    return e.toString().replace(exceptionString, "");
  }
}

/**
 * Called to either approve or deny the selected change
 * @param params collection of parameters containing the operation (Approve or Deny), the changeID that will be affected and 
 * a message to annotate the record with
 * @return messages returned from the service request otherwise a String containing the exception message
 */
function approveDenyChange(params)
{
  var sm_user = params.remove("smUser");
  var sm_password = params.remove("smPass");
  var client = new ServiceManagerScriptObject(smUrl, sm_user, sm_password);
  var operation = params.get("operation")
  var changeID = params.get("changeID");
  var message = params.get("message");
  
  try
  {
	if (operation == "Deny")
	{
      return client.denyChange(changeID, message);
	}
	else
	{
      return client.approveChange(changeID, message);
	}
  }
  catch(e)
  {
    return e.toString().replace(exceptionString, "");
  }
}

/**
 * Used to validate a user SM credentials
 *
 * Returns: collection of groups the user belongs to
 */
function validateCreds(params)
{
  var sm_user = params.remove("smUser");
  var sm_password = params.remove("smPass");
  var client = new ServiceManagerScriptObject(smUrl, sm_user, sm_password);
  var operation = params.get("operation")
  var query = params.get("query");
  try{
    return client.validateCredentials(operation, query);
  }catch(e){
    return e.toString().replace(exceptionString, "");
  }
}


var PREAMBLE = "Agent Client: HP SM CM - "
function logDebug(msg)
{
    ServiceAPI.getLogger().debug(PREAMBLE + msg);
} 

function logInfo(msg)
{
    ServiceAPI.getLogger().info(PREAMBLE + msg);
} 

function logError(msg)
{
    ServiceAPI.getLogger().error(PREAMBLE + msg);
} 
